"""Rich Printer."""

import logging
from datetime import datetime

from beartype import beartype
from beartype.typing import Any
from rich.console import Console
from rich.text import Text

from .styles import STYLES, get_name


@beartype
def rich_printer(  # noqa: CAC001
    message: str,
    *,
    is_header: bool,
    _this_level: int,
    _is_text: bool,
    # Logger-specific parameters that need to be initialized with partial(...)
    _console: Console,
    **kwargs: Any,
) -> None:
    """Generic log writer.."""
    text = Text()
    if _is_text:
        if is_header:
            print('')  # noqa: T201
        mesage_style = ('underline ' if is_header else '') + STYLES.message
        text.append(f'{message}', style=mesage_style)
    else:
        timestamp = kwargs.pop('timestamp', datetime.now())  # noqa: DTZ005
        text.append(f'{timestamp: <28} ', style=STYLES.timestamp)
        text.append('[', style=STYLES.timestamp)
        level_style = STYLES.get_style(level=_this_level)
        text.append(f'{get_name(level=_this_level): <7}', style=level_style)
        text.append(']', style=STYLES.timestamp)
        text.append(f' {message}', style=STYLES.message)

    full_lines = []
    _keys_on_own_line = kwargs.pop('_keys_on_own_line', [])
    for key in _keys_on_own_line:
        line = kwargs.pop(key, None)
        if line:
            full_lines.append((key, line))
    for key, value in kwargs.items():
        text.append(f' {key}=', style=STYLES.key)
        text.append(f'{str(value)}', style=STYLES.value)
    _console.print(text)
    for key, line in full_lines:
        new_text = Text()
        new_text.append(f' ∟ {key}', style=STYLES.key)
        new_text.append(f': {line}', style=STYLES.value_own_line)
        _console.print(new_text)

    if _this_level == logging.CRITICAL:
        _console.print_exception(show_locals=True)
        # > or 'from rich.traceback import install; install(show_locals=True)'



"""Track delegated processes with rich progress meters.

Based on: https://www.deanmontgomery.com/2022/03/24/rich-progress-and-multiprocessing

"""

import math
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from multiprocessing.managers import DictProxy
from time import sleep
from typing import runtime_checkable

from beartype import beartype
from beartype.typing import Any, List, Protocol, TypeVar, Union
from rich.progress import BarColumn, Progress, ProgressColumn, TimeElapsedColumn, TimeRemainingColumn

_ItemT = TypeVar('_ItemT', bound=Any)
"""Iterated item in the data."""


@runtime_checkable
class _DelegatedTask(Protocol):
    """Defined the kwargs accepted for a delegated task."""

    def __call__(
        self,
        *,
        task_id: int,
        shared_progress: DictProxy,  # type: ignore[type-arg]
        data: List[_ItemT],
    ) -> Any:
        ...


@beartype
def _chunked(data: List[_ItemT], count: int) -> List[List[_ItemT]]:
    """Chunk the list of data into equally sized lists."""
    # TODO: See below link for other options for chunking
    #   https://realpython.com/how-to-split-a-python-list-into-chunks/
    size = len(data)
    chunk_size, chunk_rem = size // count, size % count  # noqa: S001
    chunk_size += int(math.ceil(chunk_rem / size))
    return [
        data[ix:ix + chunk_size] for ix in range(0, size, chunk_size)
    ]


@beartype
def pretty_process(delegated_task: _DelegatedTask, *, data: List[_ItemT], num_workers: int = 3) -> Any:
    """Run a task in parallel to process all provided data.

    Uses `rich` to display pretty progress bars

    Args:
        delegated_task: must call `shared_progress[task_id] += 1` on each item in data
        data: the list of data to distribute
        num_workers: number of worker processes

    """
    # Docs: https://rich.readthedocs.io/en/latest/progress.html
    columns: List[Union[str, ProgressColumn]] = [
        '[progress.description]{task.description}',
        BarColumn(),
        '[progress.percentage]{task.percentage:>3.0f}%',
        TimeRemainingColumn(),
        TimeElapsedColumn(),
    ]
    with Progress(*columns, refresh_per_second=1) as progress:  # noqa: SIM117 (Py>3.9)
        # Share state between process and workers
        with multiprocessing.Manager() as manager:
            shared_progress = manager.dict()
            jobs = []
            totals = {}
            task_id_all = progress.add_task('[green]All jobs progress:')

            with ProcessPoolExecutor(max_workers=num_workers) as executor:
                for ix, chunk in enumerate(_chunked(data, count=num_cpus)):
                    task_id = progress.add_task(f'task {ix}')
                    shared_progress[task_id] = 0
                    totals[task_id] = len(chunk)
                    jobs.append(executor.submit(
                        delegated_task, task_id=task_id, shared_progress=shared_progress, data=chunk,
                    ))

                # Update progress bar from shared state
                remaining = len(jobs)
                while remaining:
                    n_done = 0
                    for task_id, latest in shared_progress.items():
                        n_done += latest
                        progress.update(task_id, completed=latest, total=totals[task_id])
                    progress.update(task_id_all, completed=n_done, total=len(data))
                    remaining = len(jobs) - sum(job.done() for job in jobs)

                # Collect results and catch and errors
                return [job.result() for job in jobs]


# Note: can't use beartype on a delegated_task & this function can't be in the if-block below
def __long_task(*, task_id: int, shared_progress: DictProxy, data: List[_ItemT]) -> Any:  # type: ignore[type-arg]
    """Example long task."""
    for _val in data:
        sleep(1)
        shared_progress[task_id] += 1
    return True


if __name__ == '__main__':
    # Run demo with: 'poetry run python -m shoal.pretty_process'

    # Resolve number of cores or specified maximum
    num_cpus = 4
    try:
        import psutil  # pyright: ignore
        num_cpus = psutil.cpu_count(logical=False)
    except Exception as exc:  # noqa: PIE786
        print(exc)  # noqa: T201

    result = pretty_process(__long_task, data=[*range(23)], num_workers=num_cpus)
    print(result)  # noqa: T201



"""Generic Log Writer."""

from beartype import beartype
from beartype.typing import Any


@beartype
def writer(
    message: str,  # noqa: ARG001
    *,
    is_header: bool,  # noqa: ARG001
    _this_level: int,
    _is_text: bool,
    # Logger-specific parameters that need to be initialized with partial(...)
    **kwargs: Any,  # noqa: ARG001
) -> None:
    """Generic log writer.."""
    raise NotImplementedError('The writer is for testing hot-swapping and has not yet been implemented')



"""Log."""

import logging
from functools import partial
from typing import runtime_checkable

from beartype import beartype
from beartype.typing import Any, Optional, Protocol
from rich.console import Console

from .loggers.rich_printer import rich_printer

DEF_LEVEL = logging.ERROR


@runtime_checkable
class LogCallable(Protocol):
    """Defined the kwargs accepted for a delegated task."""

    @beartype
    def __call__(
        self,
        message: str,
        *,
        is_header: bool,
        _this_level: int,
        _is_text: bool,
    ) -> Any:
        """Type-checked arguments."""


class _LogSingleton:
    """Store pointer to log function."""

    _logger: Optional[LogCallable] = None
    _log_level: int = DEF_LEVEL

    @beartype
    def set_logger(
        self,
        *,
        log_level: int,
        logger: Optional[LogCallable] = None,
        **kwargs: Any,
    ) -> LogCallable:
        """Set the internal logger instance."""
        _logger = logger or self._logger
        if not _logger:
            _logger = partial(rich_printer, _console=kwargs.setdefault('_console', Console()))
        self._logger = partial(_logger, **kwargs)
        self._log_level = log_level
        return self._logger

    @beartype
    def log(self, *args: Any, _this_level: int, is_header: bool = False, _is_text: bool = False, **kwargs: Any) -> None:
        """Delegate the arguments to the logger if this level above the threshold."""
        if _this_level < self._log_level:
            return
        # Ensure logger is configured
        _logger = self._logger or self.set_logger(log_level=self._log_level)
        _logger(*args, _this_level=_this_level, is_header=is_header, _is_text=_is_text, **kwargs)


_LOG_SINGLETON = _LogSingleton()


class _Logger:

    @beartype
    def text(self, message: str, *, is_header: bool = False, **kwargs: Any) -> None:
        """Print the content without a leading timestamp.

        If writing to a file or not natively supported by the logger, will appear in the logs as level info.

        """
        self.info(message, **{'_is_text': True, 'is_header': is_header, **kwargs})

    @beartype
    def text_debug(self, message: str, *, is_header: bool = False, **kwargs: Any) -> None:
        """Variation on text that will appear as a debug log if not supported."""
        self.debug(message, **{'_is_text': True, 'is_header': is_header, **kwargs})

    @beartype
    def debug(self, message: str, **kwargs: Any) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.DEBUG, **kwargs)

    @beartype
    def info(self, message: str, **kwargs: Any) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.INFO, **kwargs)

    @beartype
    def warning(self, message: str, **kwargs: Any) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.WARNING, **kwargs)

    @beartype
    def error(self, message: str, **kwargs: Any) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.ERROR, **kwargs)

    @beartype
    def exception(self, message: str, **kwargs: Any) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.CRITICAL, **kwargs)


@beartype
def configure_logger(*, log_level: int = DEF_LEVEL, logger: Optional[LogCallable] = None, **kwargs: Any) -> None:
    """Configure the global log level or replace the logger."""
    _LOG_SINGLETON.set_logger(logger=logger, log_level=log_level, **kwargs)


@beartype
def get_logger() -> _Logger:
    """Retrieve global logger."""
    return _Logger()


logger = get_logger()



"""Run shell commands."""

import subprocess  # noqa: S404  # nosec
import sys
from io import BufferedReader
from pathlib import Path
from time import time

from beartype import beartype
from beartype.typing import Callable, Optional

from .log import logger


@beartype
def capture_shell(
    cmd: str, *, timeout: int = 120, cwd: Optional[Path] = None, printer: Optional[Callable[[str], None]] = None,
) -> str:
    """Run shell command and return the output.

    Inspired by: https://stackoverflow.com/a/38745040/3219667

    Args:
        cmd: shell command
        timeout: process timeout. Defaults to 2 minutes
        cwd: optional path for shell execution
        printer: optional callable to output the lines in real time

    Returns:
        str: stripped output

    Raises:
        CalledProcessError: if return code is non-zero

    """
    logger.debug('Running', cmd=cmd, timeout=timeout, cwd=cwd, printer=printer)

    start = time()
    lines = []
    with subprocess.Popen(  # noqa: DUO116  # nosec  # nosemgrep
        cmd, cwd=cwd,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True,
        shell=True,  # noqa: S602
    ) as proc:
        stdout: BufferedReader = proc.stdout  # type: ignore[assignment]
        return_code = None
        while return_code is None:
            if timeout != 0 and time() - start >= timeout:
                proc.kill()
                break
            if line := stdout.readline():
                lines.append(line)
                if printer:
                    printer(line.rstrip())  # type: ignore[arg-type]
            else:
                return_code = proc.poll()

    output = ''.join(lines)  # type: ignore[arg-type]
    if return_code != 0:
        raise subprocess.CalledProcessError(returncode=return_code or 404, cmd=cmd, output=output)
    return output


@beartype
def run_shell(cmd: str, *, timeout: int = 120, cwd: Optional[Path] = None) -> None:
    """Run shell command with buffering output.

    Args:
        cmd: shell command
        timeout: process timeout. Defaults to 2 minutes
        cwd: optional path for shell execution

    Raises:
        CalledProcessError: if return code is non-zero

    """
    logger.debug('Running', cmd=cmd, timeout=timeout, cwd=cwd)

    subprocess.run(  # noqa: DUO116
        cmd, timeout=timeout or None, cwd=cwd,
        stdout=sys.stdout, stderr=sys.stderr, check=True,
        shell=True,  # noqa: S602,  # nosec  # nosemgrep
    )



"""Start the command line program."""

from beartype import beartype

from . import __pkg_name__, __version__


@beartype
def start() -> None:
    """Run the customized Invoke Program."""
    from calcipy.cli import start_program
    from calcipy.tasks import all_tasks
    start_program(__pkg_name__, __version__, all_tasks)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['corallium', 'corallium.loggers']

package_data = \
{'': ['*']}

install_requires = \
['beartype>=0.12.0', 'pydantic>=1.10.4', 'rich>=12.6.0']

extras_require = \
{':python_version < "3.11"': ['tomli>=2.0.1']}

setup_kwargs = {
    'name': 'corallium',
    'version': '0.1.0rc3',
    'description': 'Shared functionality for calcipy-ecosystem',
    'long_description': '# corallium\n\nShared functionality for the calcipy-ecosystem.\n\n## Installation\n\n1. `poetry add corallium`\n\n1. Take advantage of the logger or other common functionality\n\n    ```sh\n    form corallium.log import logger\n\n    logger.info(\'Hello!\')\n    ```\n\n## Usage\n\n<!-- < TODO: Show an example (screenshots, terminal recording, etc.) >\n\n- **log**: TBD\n- **pretty_process**: TBD\n- **shell**: TBD\n- **file_helpers**: TBD\n- **tomllib**: This is a lightweight wrapper to backport `tomli` in place of `tomllib` until we can use Python >3.11. Use with `from corallium.tomllib import tomllib`\n- **dot_dict**: has one function `ddict`, which is a light-weight wrapper around whatever is the most [maintained dotted-dictionary package in Python](https://pypi.org/search/?q=dot+accessible+dictionary&o=). Dotted dictionaries can sometimes improve code readability, but they aren\'t a one-size fits all solution. Sometimes `attr.s` or `dataclass` are more appropriate.\n    - The benefit of this wrapper is a stable interface that can be replaced with better internal implementations, such [Bunch](https://pypi.org/project/bunch/), [Chunk](https://pypi.org/project/chunk/), [Munch](https://pypi.org/project/munch/), [flexible-dotdict](https://pypi.org/project/flexible-dotdict/), [classy-json](https://pypi.org/project/classy-json/), and now [Python-Box](https://pypi.org/project/python-box/)\n -->\n\nFor more example code, see the [scripts] directory or the [tests].\n\n## Project Status\n\nSee the `Open Issues` and/or the [CODE_TAG_SUMMARY]. For release history, see the [CHANGELOG].\n\n## Contributing\n\nWe welcome pull requests! For your pull request to be accepted smoothly, we suggest that you first open a GitHub issue to discuss your idea. For resources on getting started with the code base, see the below documentation:\n\n- [DEVELOPER_GUIDE]\n- [STYLE_GUIDE]\n\n## Code of Conduct\n\nWe follow the [Contributor Covenant Code of Conduct][contributor-covenant].\n\n### Open Source Status\n\nWe try to reasonably meet most aspects of the "OpenSSF scorecard" from [Open Source Insights](https://deps.dev/pypi/corallium)\n\n## Responsible Disclosure\n\nIf you have any security issue to report, please contact the project maintainers privately. You can reach us at [dev.act.kyle@gmail.com](mailto:dev.act.kyle@gmail.com).\n\n## License\n\n[LICENSE]\n\n[changelog]: ./docs/CHANGELOG.md\n[code_tag_summary]: ./docs/CODE_TAG_SUMMARY.md\n[contributor-covenant]: https://www.contributor-covenant.org\n[developer_guide]: ./docs/DEVELOPER_GUIDE.md\n[license]: https://github.com/kyleking/corallium/LICENSE\n[scripts]: https://github.com/kyleking/corallium/scripts\n[style_guide]: ./docs/STYLE_GUIDE.md\n[tests]: https://github.com/kyleking/corallium/tests\n',
    'author': 'Kyle King',
    'author_email': 'dev.act.kyle@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyleking/corallium',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.12,<4.0.0',
}


setup(**setup_kwargs)



"""corallium."""

__version__ = '0.1.0rc3'
__pkg_name__ = 'corallium'

# ====== Above is the recommended code from calcipy_template and may be updated on new releases ======

from os import environ
from warnings import filterwarnings

from beartype.roar import BeartypeDecorHintPep585DeprecationWarning

if not environ.get('BEARTYPE_SHOW_WARNINGS'):
    # FYI: https://github.com/beartype/beartype#are-we-on-the-worst-timeline
    filterwarnings('ignore', category=BeartypeDecorHintPep585DeprecationWarning)
